import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author David Unterweger 1127540
 * 
 * 
 */
public abstract class AbsPurifier implements Prettifier {

	private final String COMMENT_REG_EXP = "(?:/\\*(?:[^*]|(?:\\*+[^*/]))*\\*+/)|(?://.*)";

	// matched jeden Zeilenumbruch inklusive dem Whitespace davor + den
	// Whitespace danach, inlusive einem evtl. vorhandenen *
	private final String ML_WHITESPACE_REG_EXP = "\\s*\n\\s*(\\*)?\\s*";

	@Override
	/*
	 * Result corresponds to the Java program in prog, but with more, less or
	 * prettier comments. This method has no side-effects. In the Result every
	 * comment in prog is in a distinct line, seperated from the Code. All
	 * comments are uniformly formatted and indented
	 */
	public String pretty(String prog) {
		// INV comments.length >= code.length
		String[] comments = getComments(prog);
		String[] code = getCode(prog);

		int commentCounter = 0;
		String prettyProg = "";

		if (!code[0].equals("")) {
			prettyProg = code[0];
		}
		for (int i = 1; i < code.length; i++) {
			int c = 1;
			while (code[i].equals("")) {
				c++;
				i++;
			}
			String lastCodeFragment = code[i];

			String comment = "";
			for (int j = 0; j < c; j++) {
				comment += decomment(comments[commentCounter++]);
			}
			prettyProg += wrapToComment(comment, getIndent(code, i - 1)) + lastCodeFragment;
		}
		return prettyProg;
	}

	// VB: s muss Zeichen enthalten die kein Whitespace sind, indent != null,
	// indent enthaelt keine Zeichen außer ' ' und '\t'
	//
	// NB: return ist ein gueltiger Java kommentar
	protected abstract String wrapToComment(String s, String indent);

	// VB:
	// comment muss ein einzelner gueltiger java Kommentar sein und darf
	// keine Nicht-Kommentar-Teile enthalten. Das heißt er muss eine der
	// folgenden Formen haben: "//einzeiligerTEXT" oder
	// "/*evtl.auch mehrzeiligerText*/"
	//
	// NB:
	// kurz: return enthaelt nur den Text aus comment, wobei jede Zeile mit \n
	// endet.
	// lang:
	// return ist der in comment enthaltene Text ohne fuehrendes //
	// ODER
	// fuehrendes /* und abschliessendes */; Ist comment mehrzeilig, sind die
	// am Beginn jeder Zeile stehenden '*' inklusive deren Einrückung und den
	// auf '*' folgenden Whitespace entfernt
	// Ausserdem ist in BEIDEN FAELLEN der Whitespace vor und nach dem Text
	// entfernt. Nachdem entfernen wurde ein "\n" angehaengt
	//
	//
	private String decomment(String comment) {
		if (comment.substring(0, 2).equals("//")) {
			return comment.substring(2).trim() + "\n";
		} else {
			if (!comment.contains("\n")) {
				return comment.substring(2, comment.length() - 2).trim() + "\n";
			} else {
				return comment.substring(2, comment.length() - 2).replaceAll(ML_WHITESPACE_REG_EXP, "\n").trim() + "\n";
			}
		}
	}

	/**
	 * VB: c < blocks.length
	 * 
	 * NB: blocks ist unveraendert. return enthaelt die Leerzeichen und
	 * Tabulatorzeichen vor dem ersten anderen Zeichen der der letzten Zeile in
	 * blocks[c], welche so ein Zeichen enthaelt. Ist keine Solche Zeile in
	 * blocks[c] wird blocks[c-1] durchsucht. enthaelt kein String in blocks
	 * Zeichen ungleich ' ' und '\t' ist return ""
	 */
	private String getIndent(String[] blocks, int c) {
		if (c < 0) {
			return "";
		}
		String[] s = blocks[c].split("\n");

		for (int i = s.length - 1; i >= 0; i--) {
			if (!(s[i].trim().equals(""))) {
				return getIndentOfLine(s[i]);
			}
		}
		return getIndent(blocks, c - 1);
	}

	/**
	 * VB: s muss mindestens ein Zeichen ungleich ' ' oder '\t' enthalten.
	 * 
	 * NB: return enthaelt genau jene Zeichen aus s, welche vor dem ersten
	 * auftreten eines Zeichens ungleich ' ' und '\t' stehen.
	 */
	private String getIndentOfLine(String s) {
		String indent = "";
		for (int i = 0; ((s.charAt(i) == ' ') || (s.charAt(i) == '\t')) && i < s.length(); i++) {
			indent += s.charAt(i);
		}
		System.out.println(indent);
		return indent;
	}

	/**
	 * VB: prog != null;
	 * 
	 * NB: return = Array aus Strings welche nur den Code darstellen. Die
	 * Einrueckung der einzelnen COdefragmente bleibt erhalten. Codefragmente
	 * welche nur Whitespace enthielten sind leere Strings, bei allen anderen
	 * Fragmenten ist der Whitespace am Ende entfernt und ein '\n' angehaengt
	 */
	private String[] getCode(String prog) {
		String[] code = prog.split("(" + COMMENT_REG_EXP + ")([\\t| ]*\\n)?");
		for (int i = 0; i < code.length; i++) {
			if (Pattern.compile("\\s*").matcher(code[i]).matches()) {
				code[i] = "";
			} else {
				code[i] = trimEnd(code[i]);
			}
		}
		return code;
	}

	/**
	 * VB: prog != null; NB: return = Array aus Strings welche die einzenen
	 * Kommentare aus prog darstellen
	 */
	private String[] getComments(String prog) {
		Matcher matcher = Pattern.compile(COMMENT_REG_EXP).matcher(prog);
		ArrayList<String> comments = new ArrayList<String>();

		while (matcher.find()) {
			comments.add(matcher.group());
		}
		return comments.toArray(new String[comments.size()]);
	}

	// VB: s!= null
	// NB: return ist s nur ohne whitespace am ende, allerdings mit
	// abschließendem '\n'
	private String trimEnd(String s) {
		return s.replaceFirst("\\s*\\z", "\n");
	}

	// VB: s muss Zeichen enthalten die kein Whitespace sind, s darf den
	// substring "*/" nicht enthalten; indent != null, indent enthaelt keine
	// Zeichen außer ' ' und '\t'
	//
	// NB: return ist ein gültiger Javakommentar der Form:
	// > indent+// Zeile1\n
	// > indent+// Zeile2\n
	// > indent+// ...\n
	protected String wrapToLine(String s, String indent) {
		String[] lines = s.trim().split("\n");
		String comment = "";
		for (int i = 0; i < lines.length; i++) {
			comment += indent + "// " + lines[i].trim() + "\n";
		}
		return comment;
	}

	// VB:
	// s muss Zeichen enthalten die kein Whitespace sind, s darf den substring
	// "*/" nicht enthalten; indent != null, indent enthaelt keine Zeichen außer
	// ' ' und '\t'
	//
	// NB:
	// return ist ein gueltiger JavaKommentar
	// falls s einzeilig, hat return folgende Form "/* TEXT */\n"
	// falls s mehrzeilig, hat return folgende Form:
	// >indent+/* ZEILE1
	// >indent+ * ZEILE2
	// >indent+ * ...
	// >indent+ */\n
	protected String wrapToMultiLine(String s, String indent) {

		if (!s.trim().contains("\n")) {
			// s ist einzeilig
			return indent + "/* " + s.trim() + " */\n";
		} else {
			// s ist mehrzeilig

			String[] lines = s.trim().split("\n");
			String comment = indent + "/* " + lines[0] + "\n";

			for (int i = 1; i < lines.length; i++) {
				comment += indent + " * " + lines[i] + "\n";
			}
			comment += indent + " */\n";
			return comment;
		}
	}

	// VB: s muss Zeichen enthalten die kein Whitespace sind, s darf den
	// substring "*/" nicht enthalten; indent != null, indent enthaelt keine
	// Zeichen außer ' ' und '\t'
	//
	// NB: return ist ein gültiger Javakommentar der Form:
	// falls s einzeilig:
	// >indent+// TEXT\n
	// falls s mehrzeilig:
	// >indent+/* Zeile1\n
	// >indent+ * Zeile2\n
	// >indent+ * ...\n
	// >indent+ */\n
	protected String wrapToLineMultiLine(String s, String indent) {
		if (!s.trim().contains("\n")) {
			// Kommentar ist einzeilig
			return indent + "// " + s.trim() + "\n";
		} else {
			// Kommentar ist mehrzeilig
			String[] lines = s.trim().split("\n");
			String comment = indent + "/* " + lines[0].trim() + "\n";
			for (int i = 1; i < lines.length; i++) {
				comment += indent + " * " + lines[i].trim() + "\n";
			}
			return comment + indent + " */\n";
		}
	}

}