import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * 
 * @author 
 *
 */
public class Composer implements Prettifier {
	// never null
	private SortedMap<Integer, Prettifier> prettifiers = new TreeMap<Integer, Prettifier>();
	
	/**
	 * prettifies the given program by iterating over all previousely stored prettifiers
	 * and applying their prettification-functions.
	 */
	@Override
	public String pretty(String prog) {
		// as none of the calling prettifictions has side effects this is ok
		String prettified = prog;
		SortedSet<Integer> keys = new TreeSet<Integer>(prettifiers.keySet());
		for (Integer key : keys) { 
			prettified = prettifiers.get(key).pretty(prettified);
		}
		return prettified;
	}
	
	/**
	 * Adds prettifier p at given position. If there is already a prettifier at the given position it will
	 * be replaced by the new one. position has to be positive
	 * @param position
	 * @param p
	 */
	public void addPrettifier(int position, Prettifier p) {
		if (position > 0)
			prettifiers.put(position, p);
	}
	
	/**
	 * Removes a prettifier from position pos if present
	 * @param pos
	 */
	public void removePrettifier(int pos) {
		prettifiers.remove(pos);
	}
}
